@@ -89,6 +89,7 @@ sub read_res_packet {
 
     my $err = sub {
         my $code = shift;
+        $sock->close() if $sock->connected;
         $$err_ref = $code if ref $err_ref;
         return undef;
     };
@@ -104,19 +105,19 @@ sub read_res_packet {
     return $err->("malformed_magic") unless $magic eq "\0RES";
 
     if ($len) {
-        # Start off trying to read the whole buffer. Store the bits in an array
-        # one element for each read, then do a big join at the end. This minimizes
-        # the number of memory allocations we have to do.
         my $readlen = $len;
+        my $offset = 0;
         my $lim = 20 + int( $len / 2**10 );
-        my @buffers;
         for (my $i = 0; $readlen > 0 && $i < $lim; $i++) {
-            my $rv = sysread($sock, $buffers[$i], $readlen);
+            # Because we know the length of the data we need to read exactly, the
+            # most efficient way to do this in perl is with one giant buffer, and
+            # an appropriate offset passed to sysread.
+            my $rv = sysread($sock, $buf, $readlen, $offset);
             return $err->("short_body") unless $rv > 0;
             last unless $rv > 0;
             $readlen -= $rv;
+            $offset += $rv;
         }
-        $buf = join('', @buffers);
         return $err->("short_body") unless length($buf) == $len; 
     }
 
